@@ -10,21 +10,17 @@
 
 
 use core::convert::TryFrom;
-use core::marker::PhantomData;
 use core::fmt::Debug;
 
+use base::{Base, SpiBase, HalError};
 use log::{trace, debug, warn};
 
 use embedded_hal::spi::{Mode as SpiMode, Phase, Polarity};
-use embedded_hal::spi::blocking::{Transfer, Write, Transactional};
-use embedded_hal::delay::blocking::{DelayMs, DelayUs};
+use embedded_hal::delay::blocking::{DelayUs};
 use embedded_hal::digital::blocking::{InputPin, OutputPin};
 
-use driver_pal::{wrapper::Wrapper as SpiWrapper, Error as WrapError};
-
 use radio::{Power as _, State as _};
 
-
 pub mod base;
 
 pub mod device;
@@ -53,12 +49,8 @@ pub enum Mode {
 ///
 /// Operating functions are implemented as traits from the [radio] package
 ///
-pub struct Sx127x<Base, CommsError, PinError, DelayError> {
-    hal: Base,
-
-    _ce: PhantomData<CommsError>,
-    _pe: PhantomData<PinError>,
-    _de: PhantomData<DelayError>,
+pub struct Sx127x<Hal> {
+    hal: Hal,
 
     mode: Mode,
     config: Config,
@@ -66,13 +58,9 @@ pub struct Sx127x<Base, CommsError, PinError, DelayError> {
 
 /// Sx127x error type
 #[derive(Debug, Clone, PartialEq)]
-pub enum Error<CommsError, PinError, DelayError> {
-    /// Communications (SPI or UART) error
-    Comms(CommsError),
-    /// Pin control error
-    Pin(PinError),
-    /// Delay error
-    Delay(DelayError),
+pub enum Error<HalError: Debug + 'static> {
+    /// Communications (SPI, UART, pin, delay) error
+    Hal(HalError),
     /// Invalid configuration
     InvalidConfiguration,
     /// Transaction aborted
@@ -89,14 +77,9 @@ pub enum Error<CommsError, PinError, DelayError> {
     InvalidDevice(u8),
 }
 
-impl<CommsError, PinError, DelayError> From<WrapError<CommsError, PinError, DelayError>> for Error<CommsError, PinError, DelayError> {
-    fn from(e: WrapError<CommsError, PinError, DelayError>) -> Self {
-        match e {
-            WrapError::Spi(e) => Error::Comms(e),
-            WrapError::Pin(e) => Error::Pin(e),
-            WrapError::Delay(e) => Error::Delay(e),
-            WrapError::Aborted => Error::Aborted,
-        }
+impl<HalError: Debug + 'static> From<HalError> for Error<HalError> {
+    fn from(e: HalError) -> Self {
+        Error::Hal(e)
     }
 }
 
@@ -114,20 +97,24 @@ impl Default for Settings {
     }
 }
 
-pub type Sx127xSpi<Spi, SpiError, CsPin, BusyPin, ReadyPin, SdnPin, PinError, Delay, DelayError> = Sx127x<SpiWrapper<Spi, CsPin, BusyPin, ReadyPin, SdnPin, Delay>, SpiError, PinError, DelayError>;
+pub type Sx127xSpi<Spi, CsPin, BusyPin, ReadyPin, SdnPin, Delay> = Sx127x<base::Base<Spi, CsPin, BusyPin, ReadyPin, SdnPin, Delay>>;
 
-impl<Spi, SpiError, CsPin, BusyPin, ReadyPin, ResetPin, PinError, Delay, DelayError>
-    Sx127x<SpiWrapper<Spi, CsPin, BusyPin, ReadyPin, ResetPin, Delay>, SpiError, PinError, DelayError>
+impl<Spi, CsPin, BusyPin, ReadyPin, SdnPin, PinError, Delay>
+    Sx127x<
+        Base<Spi, CsPin, BusyPin, ReadyPin, SdnPin, Delay>,
+    >
 where
-    Spi: Transfer<u8, Error = SpiError> + Write<u8, Error = SpiError> + Transactional<u8, Error = SpiError>,
+    Spi: SpiBase,
+    <Spi as SpiBase>::Error: Debug,
+
     CsPin: OutputPin<Error = PinError>,
     BusyPin: InputPin<Error = PinError>,
     ReadyPin: InputPin<Error = PinError>,
-    ResetPin: OutputPin<Error = PinError>,
-    Delay: DelayMs<u32, Error = DelayError> + DelayUs<u32, Error = DelayError>,
-    SpiError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    SdnPin: OutputPin<Error = PinError>,
+    PinError: Debug,
+
+    Delay: DelayUs,
+    <Delay as DelayUs>::Error: Debug,
 {
     /// Create an Sx127x with the provided SPI implementation and pins
     pub fn spi(
@@ -135,27 +122,24 @@ where
         cs: CsPin,
         busy: BusyPin,
         ready: ReadyPin,
-        reset: ResetPin,
+        sdn: SdnPin,
         delay: Delay,
         config: &Config,
-    ) -> Result<Self, Error<SpiError, PinError, DelayError>> {
+    ) -> Result<Self, Error<HalError<<Spi as SpiBase>::Error, PinError, <Delay as DelayUs>::Error>>> {
         // Create SpiWrapper over spi/cs/busy/ready/reset
-        let hal = SpiWrapper::new(spi, cs, reset, busy, ready, delay);
+        let base = Base{spi, cs, sdn, busy, ready, delay};
 
         // Create instance with new hal
-        Self::new(hal, config)
+        Self::new(base, config)
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal,
 {
     /// Create a new radio instance over an arbitrary base::Base implementation
-    pub fn new(hal: Base, config: &Config) -> Result<Self, Error<CommsError, PinError, DelayError>> {
+    pub fn new(hal: Hal, config: &Config) -> Result<Self, Error<<Hal as base::Hal>::Error>> {
         // Build container object
         let mut sx127x = Self::build(hal, config.clone());
 
@@ -187,7 +171,7 @@ where
 
     /// Reset the device. This recalibrates and patches the device,
     /// however, still requires configuration.
-    pub fn reset(&mut self) -> Result<(), Error<CommsError, PinError, DelayError>> {
+    pub fn reset(&mut self) -> Result<(), Error<<Hal as base::Hal>::Error>> {
         self.hal.reset()?;
 
         // Calibrate RX chain
@@ -206,7 +190,7 @@ where
     }
 
     /// (re)apply device configuration
-    pub fn configure(&mut self, config: &Config) -> Result<(), Error<CommsError, PinError, DelayError>> {
+    pub fn configure(&mut self, config: &Config) -> Result<(), Error<<Hal as base::Hal>::Error>> {
         // Configure the modem as appropriate
         match (&config.modem, &config.channel) {
             (Modem::LoRa(lora_modem), Channel::LoRa(lora_channel)) => {
@@ -228,7 +212,7 @@ where
     }
 
     /// Fetch device silicon version
-    pub fn silicon_version(&mut self) -> Result<u8, Error<CommsError, PinError, DelayError>> {
+    pub fn silicon_version(&mut self) -> Result<u8, Error<<Hal as base::Hal>::Error>> {
         self.read_reg(regs::Common::VERSION)
     }
 
@@ -236,7 +220,7 @@ where
     pub(crate) fn configure_pa(
         &mut self,
         pa_config: &PaConfig,
-    ) -> Result<(), Error<CommsError, PinError, DelayError>> {
+    ) -> Result<(), Error<<Hal as base::Hal>::Error>> {
         use device::*;
 
         // Set output configuration
@@ -261,7 +245,7 @@ where
     pub(crate) fn set_state_checked(
         &mut self,
         state: State,
-    ) -> Result<(), Error<CommsError, PinError, DelayError>> {
+    ) -> Result<(), Error<<Hal as base::Hal>::Error>> {
         // Send set state command
         trace!("Set state to: {:?} (0x{:02x})", state, state as u8);
         self.set_state(state)?;
@@ -283,7 +267,7 @@ where
             }
             
 
-            self.hal.delay_ms(1).map_err(Error::Delay)?;
+            self.hal.delay_ms(1)?;
             ticks += 1;
         }
         Ok(())
@@ -306,7 +290,7 @@ where
     pub(crate) fn set_modem(
         &mut self,
         modem: ModemMode,
-    ) -> Result<(), Error<CommsError, PinError, DelayError>> {
+    ) -> Result<(), Error<<Hal as base::Hal>::Error>> {
         match modem {
             ModemMode::Standard => {
                 self.set_state(State::Sleep)?;
@@ -330,7 +314,7 @@ where
     }
 
     // Set the channel by frequency
-    pub(crate) fn set_frequency(&mut self, freq: u32) -> Result<(), Error<CommsError, PinError, DelayError>> {
+    pub(crate) fn set_frequency(&mut self, freq: u32) -> Result<(), Error<<Hal as base::Hal>::Error>> {
         let channel = self.freq_to_channel_index(freq);
 
         let outgoing = [
@@ -347,7 +331,7 @@ where
     }
 
     // Fetch the current channel index
-    pub(crate) fn get_frequency(&mut self) -> Result<u32, Error<CommsError, PinError, DelayError>> {
+    pub(crate) fn get_frequency(&mut self) -> Result<u32, Error<<Hal as base::Hal>::Error>> {
         let mut incoming = [0u8; 3];
 
         self.hal
@@ -361,7 +345,7 @@ where
 
     /// Calibrate the device RF chain
     /// This MUST be called directly after resetting the module
-    pub(crate) fn rf_chain_calibration(&mut self) -> Result<(), Error<CommsError, PinError, DelayError>> {
+    pub(crate) fn rf_chain_calibration(&mut self) -> Result<(), Error<<Hal as base::Hal>::Error>> {
         debug!("Running calibration");
 
         // Load initial PA config
@@ -410,48 +394,36 @@ where
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal,
 {
-    pub(crate) fn build(hal: Base, config: Config) -> Self {
+    pub(crate) fn build(hal: Hal, config: Config) -> Self {
         Sx127x {
             hal,
             config,
             mode: Mode::Unconfigured,
-            _ce: PhantomData,
-            _pe: PhantomData,
-            _de: PhantomData,
         }
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> DelayMs<u32> for Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> DelayUs for Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal,
 {
-    type Error = Error<CommsError, PinError, DelayError>;
+    type Error = Error<<Hal as base::Hal>::Error>;
 
-    fn delay_ms(&mut self, t: u32) -> Result<(), Error<CommsError, PinError, DelayError>> {
-        self.hal.delay_ms(t).map_err(Error::Delay)
+    fn delay_us(&mut self, t: u32) -> Result<(), Error<<Hal as base::Hal>::Error>> {
+        self.hal.delay_us(t).map_err(Error::Hal)
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal,
 {
     /// Read a u8 value from the specified register
-    pub fn read_reg<R>(&mut self, reg: R) -> Result<u8, Error<CommsError, PinError, DelayError>>
+    pub fn read_reg<R>(&mut self, reg: R) -> Result<u8, Error<<Hal as base::Hal>::Error>>
     where
         R: Copy + Clone + Debug + Into<u8>,
     {
@@ -463,13 +435,13 @@ where
     }
 
     /// Write a u8 value to the specified register
-    pub fn write_reg<R>(&mut self, reg: R, value: u8) -> Result<(), Error<CommsError, PinError, DelayError>>
+    pub fn write_reg<R>(&mut self, reg: R, value: u8) -> Result<(), Error<<Hal as base::Hal>::Error>>
     where
         R: Copy + Clone + Debug + Into<u8>,
     {
         trace!("Write reg:  {:?} (0x{:02x}): 0x{:02x}", reg, reg.into(), value);
 
-        self.hal.write_reg(reg.into(), value)
+        self.hal.write_reg(reg.into(), value).map_err(Error::Hal)
     }
 
     /// Update the specified register with the provided (value & mask)
@@ -478,25 +450,22 @@ where
         reg: R,
         mask: u8,
         value: u8,
-    ) -> Result<u8, Error<CommsError, PinError, DelayError>>
+    ) -> Result<u8, Error<<Hal as base::Hal>::Error>>
     where
         R: Copy + Clone + Debug + Into<u8>,
     {
         trace!("Update reg: {:?} (0x{:02x}): 0x{:02x} (0x{:02x})", reg, reg.into(), value, mask);
 
-        self.hal.update_reg(reg.into(), mask, value)
+        self.hal.update_reg(reg.into(), mask, value).map_err(Error::Hal)
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> radio::State for Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> radio::State for Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal
 {
     type State = State;
-    type Error = Error<CommsError, PinError, DelayError>;
+    type Error = Error<<Hal as base::Hal>::Error>;
 
     /// Fetch device state
     fn get_state(&mut self) -> Result<Self::State, Self::Error> {
@@ -513,17 +482,14 @@ where
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> radio::Power for Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> radio::Power for Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal
 {
-    type Error = Error<CommsError, PinError, DelayError>;
+    type Error = Error<<Hal as base::Hal>::Error>;
 
     /// Set transmit power (using the existing `PaConfig`)
-    fn set_power(&mut self, power: i8) -> Result<(), Error<CommsError, PinError, DelayError>> {
+    fn set_power(&mut self, power: i8) -> Result<(), Error<<Hal as base::Hal>::Error>> {
         use device::*;
 
         // Limit to viable input range
@@ -565,15 +531,12 @@ where
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> radio::Interrupts for Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> radio::Interrupts for Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal
 {
     type Irq = Interrupts;
-    type Error = Error<CommsError, PinError, DelayError>;
+    type Error = Error<<Hal as base::Hal>::Error>;
 
     /// Fetch pending interrupts from the device
     /// If the clear option is set, this will also clear any pending flags
@@ -586,18 +549,15 @@ where
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> radio::Channel for Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> radio::Channel for Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal
 {
     type Channel = Channel;
-    type Error = Error<CommsError, PinError, DelayError>;
+    type Error = Error<<Hal as base::Hal>::Error>;
 
     /// Set the LoRa mode channel for future receive or transmit operations
-    fn set_channel(&mut self, channel: &Self::Channel) -> Result<(), Error<CommsError, PinError, DelayError>> {
+    fn set_channel(&mut self, channel: &Self::Channel) -> Result<(), Error<<Hal as base::Hal>::Error>> {
         match (self.mode, channel) {
             (Mode::LoRa, Channel::LoRa(channel)) => self.lora_set_channel(channel),
             (Mode::FskOok, Channel::FskOok(channel)) => self.fsk_set_channel(channel),
@@ -606,14 +566,11 @@ where
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> radio::Transmit for Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> radio::Transmit for Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal
 {
-    type Error = Error<CommsError, PinError, DelayError>;
+    type Error = Error<<Hal as base::Hal>::Error>;
 
     /// Start sending a packet
     fn start_transmit(&mut self, data: &[u8]) -> Result<(), Self::Error> {
@@ -627,7 +584,7 @@ where
     /// Check for transmission completion
     /// This method should be polled (or checked following and interrupt) to indicate sending
     /// has completed
-    fn check_transmit(&mut self) -> Result<bool, Error<CommsError, PinError, DelayError>> {
+    fn check_transmit(&mut self) -> Result<bool, Error<<Hal as base::Hal>::Error>> {
         match self.mode {
             Mode::LoRa => self.lora_check_transmit(),
             Mode::FskOok => self.fsk_check_transmit(),
@@ -636,15 +593,12 @@ where
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> radio::Receive for Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> radio::Receive for Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal
 {
     type Info = PacketInfo;
-    type Error = Error<CommsError, PinError, DelayError>;
+    type Error = Error<<Hal as base::Hal>::Error>;
 
     /// Start receive mode
     fn start_receive(&mut self) -> Result<(), Self::Error> {
@@ -674,29 +628,25 @@ where
     ///  and returns the number of bytes received on success
     fn get_received(
         &mut self,
-        info: &mut Self::Info,
-        data: &mut [u8],
-    ) -> Result<usize, Self::Error> {
+        buff: &mut [u8],
+    ) -> Result<(usize, Self::Info), Self::Error> {
         match self.mode {
-            Mode::LoRa => self.lora_get_received(info, data),
-            Mode::FskOok => self.fsk_get_received(info, data),
+            Mode::LoRa => self.lora_get_received(buff),
+            Mode::FskOok => self.fsk_get_received(buff),
             _ => Err(Error::InvalidConfiguration),
         }
     }
 }
 
-impl<Base, CommsError, PinError, DelayError> radio::Rssi for Sx127x<Base, CommsError, PinError, DelayError>
+impl<Hal> radio::Rssi for Sx127x<Hal>
 where
-    Base: base::Base<CommsError, PinError, DelayError>,
-    CommsError: Debug + Sync + Send + 'static,
-    PinError: Debug + Sync + Send + 'static,
-    DelayError: Debug + Sync + Send + 'static,
+    Hal: base::Hal
 {
-    type Error = Error<CommsError, PinError, DelayError>;
+    type Error = Error<<Hal as base::Hal>::Error>;
 
     /// Poll for the current channel RSSI
     /// This should only be called in receive mode
-    fn poll_rssi(&mut self) -> Result<i16, Error<CommsError, PinError, DelayError>> {
+    fn poll_rssi(&mut self) -> Result<i16, Error<<Hal as base::Hal>::Error>> {
         match self.mode {
             Mode::LoRa => self.lora_poll_rssi(),
             Mode::FskOok => self.fsk_poll_rssi(),
